# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from django.utils.html import escape
from django.utils.encoding import force_str


class SelectFormatter(object):
    """
    Node formatter in conjunction with `fields.TreeSelect`
    and `fields.TreeSelectMultiple` fields.
    The formatter gets initialized with the attribute name of
    the current field, the selected values (pk values normally),
    disabled pks and the field settings.

    NOTE: Always build the HTML ids with the template string
    `ID_TEMPLATE`, otherwise javascript will not recognize the
    ids correctly (Hardcoded to `'treewidget_<attr_name>_<pk value>'`)
    """
    ID_TEMPLATE = 'treewidget_%s_%s'

    def __init__(self, attr_name, selected, disabled, settings):
        self.attr_name = attr_name
        self.selected = selected
        self.disabled = disabled
        self.settings = settings or {}

    def render(self, queryset):
        """
        Render method of tree data for `jstree`.
        NOTE: `queryset` is a `tree.TreeQueryset` object.
        To avoid expensive database lookups, the parent pk
        is accessible as `node.node._parent_pk`.
        """
        for node in queryset:
            id = self.ID_TEMPLATE % (self.attr_name, node.pk)
            parent = '#'
            #try:
            if node.node._parent_pk:
                parent = self.ID_TEMPLATE % (self.attr_name, node.node._parent_pk)
            #except AttributeError:
            #    parent_obj = node.parent
            #    if parent_obj:
            #        parent = self.ID_TEMPLATE % (self.attr_name, parent_obj.node.pk)
            yield {
                'id': id,
                'parent': parent,
                'text': escape(force_str(node)),
                'data': {
                    'sort': node.ordering if self.settings.get('sort') else []
                },
                'state': {
                    'selected': True if str(node.pk) in self.selected else False,
                    'disabled': True if node.pk in self.disabled else False
                }
            }



from django.forms.widgets import SelectMultiple, Select
from django.forms import ModelChoiceField, ModelMultipleChoiceField
from django.utils.safestring import mark_safe
from django.db import models
from django.conf import settings
from json import dumps
from django.urls import reverse, NoReverseMatch
from treewidget.tree import TreeQuerySet, get_treetype, MPTT
from treewidget.formatters import SelectFormatter


TREEOPTIONS = {
    'plugins': ['wholerow'],
    'core': {
        'themes': {'icons': False, 'dots': False},
        'check_callback': True,
        'multiple': True,
        'animation': False
    },
    'search': {'show_only_matches': True}
}


class TreeSelectWidgetMixin(object):
    """
    Mixin class for SelectWidgets to provide the tree functionality.
    """
    settings = {}
    treeoptions = ''
    multiple = False
    choices = None

    class Media:
        js = (
            'treewidget/jstree.min.js',
            'treewidget/default.js',
        )
        css = {'screen': (
            'treewidget/themes/default/style.css',
            'treewidget/default.css'
        )}

    def prepare_queryset(self, selected):
        """
        Prepares the underlying queryset so it can be used for the jstree
        data generation in the formatter.

        Steps taken:
            - convert selected to a list of pks (as strings)
            - annotate _parent_pk to objects to avoid db query for parent lookup
            - replace `choices` with drilled down list to avoid another db query
            - if `filtered` in settings is `True` replace queryset with
              queryset containing all ancestors to ensure the data form
              a correct subtree structure (disables added nodes in treewidget)
        """
        # set selected to a list of str(pk)
        if not selected:
            selected = []
        elif isinstance(selected, str):
            selected = [selected]
        elif not hasattr(selected, '__iter__'):
            selected = [selected]
        selected = [str(pk) for pk in selected]

        # add _parent_pk attribute to queryset objects
        qs = TreeQuerySet(self.choices.queryset).annotate_parent()

        # replace choices to avoid another db query
        # FIXME: has illegal state in 2nd run, therefore disabled for now
        # FIXME: also causes trouble with `empty_label` (
        #choices = []
        #if self.settings.get('empty_label'):
        #    choices.append(('', self.settings['empty_label']))
        #for node in qs:
        #    choices.append((node.pk, force_text(node)))
        #self.choices = choices

        if not self.settings.get('filtered'):
            return qs, selected, []

        orig_pks = set(node.pk for node in qs)
        qs_new = TreeQuerySet(qs).get_ancestors_parent_annotated(include_self=True)
        disabled = set(node.pk for node in qs_new) - orig_pks
        return qs_new, selected, disabled

    def _get_mixin_context(self, name, qs, selected, disabled, attrs=None):
        """
        Method to build the final tree widget context data.
        The tree data is provided to `jstree` as json object in the DOM.
        The data for `jstree` rendered by `formatters.SelectFormatter`.
        """
        # need something like a unique id, use name if none in attrs
        if not attrs or not attrs.get('id'):
            attrs = {'id': name}
        attr_name = attrs.get('id')

        # try to get ajax urls
        try:
            update_url = reverse('treewidget.get_node')
            move_url = reverse('treewidget.move_node')
        except NoReverseMatch:
            update_url = ''
            move_url = ''

        # load settings if not supplied
        if not self.settings and hasattr(settings, 'TREEWIDGET_SETTINGS'):
            self.settings = settings.TREEWIDGET_SETTINGS
        if not self.treeoptions:
            self.treeoptions = dumps(settings.TREEWIDGET_TREEOPTIONS
                if hasattr(settings, 'TREEWIDGET_TREEOPTIONS') else TREEOPTIONS)

        # jstree data formatter
        formatter = (self.settings.get('formatter') or SelectFormatter)(
            attr_name, selected, disabled, self.settings)

        # additional settings for JS
        additional = {
            'id': attr_name,
            'appmodel': qs.appmodel,
            'disabled': attrs.get('disabled', False),
            'multiple': self.multiple,
            'search': self.settings.get('search', False),
            'show_buttons': self.settings.get('show_buttons', False),
            'sort': qs.ordering_signs if self.settings.get('sort') else [],
            'updateurl': update_url,
            'dnd': self.settings.get('dnd', False),
            'moveurl': move_url if self.settings.get('dnd') else '',
        }

        # data for JS
        json_data = '{"settings": %s, "additional": %s, "treedata": %s}' % (
            self.treeoptions,
            dumps(additional),
            dumps(list(formatter.render(qs)))
        )

        # treewidget context
        return {
            'id': attr_name,
            'search': self.settings.get('search', False),
            'show_buttons': self.settings.get('show_buttons', False),
            'json_data': mark_safe(json_data.replace('</script', '<\/script')),
            'disabled': attrs.get('disabled')
        }


class TreeSelectMultiple(SelectMultiple, TreeSelectWidgetMixin):
    template_name = 'treewidget/treewidget.html'
    multiple = True

    def get_context(self, name, value, attrs):
        drawable_qs, selected, disabled = self.prepare_queryset(value)
        ctx = super(TreeSelectMultiple, self).get_context(name, value, attrs)
        ctx['widget']['treewidget'] = self._get_mixin_context(
            name, drawable_qs, selected, disabled, attrs)
        ctx['widget']['treewidget']['super_template'] = super(TreeSelectMultiple, self).template_name
        return ctx


class TreeSelect(Select, TreeSelectWidgetMixin):
    template_name = 'treewidget/treewidget.html'
    multiple = False

    def get_context(self, name, value, attrs):
        drawable_qs, selected, disabled = self.prepare_queryset(value)
        ctx = super(TreeSelect, self).get_context(name, value, attrs)
        ctx['widget']['treewidget'] = self._get_mixin_context(
            name, drawable_qs, selected, disabled, attrs)
        ctx['widget']['treewidget']['super_template'] = super(TreeSelect, self).template_name
        return ctx


class TreeModelFieldMixin(object):
    def __init__(self, queryset, *args, **kwargs):
        if hasattr(queryset, 'model') and get_treetype(queryset.model) == MPTT:
            mptt_opts = queryset.model._mptt_meta
            queryset = queryset.order_by(mptt_opts.tree_id_attr, mptt_opts.left_attr)
        super(TreeModelFieldMixin, self).__init__(queryset, *args, **kwargs)


class TreeModelMultipleChoiceField(TreeModelFieldMixin, ModelMultipleChoiceField):
    widget = TreeSelectMultiple

    def __init__(self, queryset, *args, **kwargs):
        self.settings = kwargs.pop('settings', {})
        self.treeoptions = kwargs.pop('treeoptions', '')
        super(TreeModelMultipleChoiceField, self).__init__(queryset, *args, **kwargs)
        #self.settings['empty_label'] = self.empty_label
        self.widget.settings = self.settings
        self.widget.treeoptions = self.treeoptions


class TreeModelChoiceField(TreeModelFieldMixin, ModelChoiceField):
    widget = TreeSelect

    def __init__(self, queryset, *args, **kwargs):
        self.settings = kwargs.pop('settings', {})
        self.treeoptions = kwargs.pop('treeoptions', '')
        super(TreeModelChoiceField, self).__init__(queryset, *args, **kwargs)
        #self.settings['empty_label'] = self.empty_label
        self.widget.settings = self.settings
        self.widget.treeoptions = self.treeoptions


class TreeForeignKey(models.ForeignKey):
    def __init__(self, *args, **kwargs):
        self.settings = kwargs.pop('settings', {})
        self.treeoptions = kwargs.pop('treeoptions', '')
        super(TreeForeignKey, self).__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        kwargs['form_class'] = TreeModelChoiceField
        kwargs['settings'] = self.settings
        kwargs['treeoptions'] = self.treeoptions
        return super(TreeForeignKey, self).formfield(**kwargs)


class TreeOneToOneField(models.OneToOneField):
    def __init__(self, *args, **kwargs):
        self.settings = kwargs.pop('settings', {})
        self.treeoptions = kwargs.pop('treeoptions', '')
        super(TreeOneToOneField, self).__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        kwargs['form_class'] = TreeModelChoiceField
        kwargs['settings'] = self.settings
        kwargs['treeoptions'] = self.treeoptions
        return super(TreeOneToOneField, self).formfield(**kwargs)


class TreeManyToManyField(models.ManyToManyField):
    def __init__(self, *args, **kwargs):
        self.settings = kwargs.pop('settings', {})
        self.treeoptions = kwargs.pop('treeoptions', '')
        super(TreeManyToManyField, self).__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        kwargs['form_class'] = TreeModelMultipleChoiceField
        kwargs['settings'] = self.settings
        kwargs['treeoptions'] = self.treeoptions
        return super(TreeManyToManyField, self).formfield(**kwargs)



from django.http import JsonResponse
from django.apps import apps
from django.contrib.auth.decorators import login_required
from django.utils.html import escape
from django.utils.encoding import force_str
from treewidget.tree import TreeQuerySet, TreeNode

# TODO: check for individual permissions


@login_required
def get_node(request):
    """
    Ajax view for node changes with the admin popup.
    Since django's popup functionality itself does not return
    enough data to keep a tree in sync, we have to request
    changes with additional tree data separately.
    Handles add and tree moves with the popup.
    Returns a list of requested nodes with their parents,
    direct siblings and sort order.
    :param request:
    :return:
    """
    appmodel = request.GET.get('appmodel', None)
    ids = request.GET.getlist('ids')
    sort = request.GET.get('sort')
    if not appmodel or not ids:
        return JsonResponse([], safe=False)
    try:
        app_label, model_name = appmodel.split('.')
        model = apps.get_model(app_label=app_label, model_name=model_name)
        result = []
        for elem in TreeQuerySet(model.objects.filter(id__in=ids)):

            # get parent nodes
            parents = [{
                'name': escape(force_str(p)),
                'parent': p.parent.node.pk if p.parent else None,
                'id': p.node.pk,
                'sort': p.ordering if sort else None
            } for p in elem.ancestors]

            result.append({
                'name': escape(force_str(elem)),
                'parent': elem.parent.node.pk if elem.parent else None,
                'id': elem.node.pk,
                'sort': elem.ordering if sort else None,
                'parents': parents,
                'prev': elem.prev_sibling.node.pk if elem.prev_sibling else None,
                'next': elem.next_sibling.node.pk if elem.next_sibling else None
            })
        return JsonResponse(result, safe=False)
    except:
        return JsonResponse([], safe=False)


@login_required
def move_node(request):
    """
    Ajax view to evaluate and execute direct drag'n drop moves in the tree.
    Returns True if the moves is legal, otherwise False.
    Not used, if drag'n drop is disabled.
    :param request:
    :return:
    """
    appmodel = request.GET.get('appmodel', None)
    node_id = request.GET.get('id', None)
    parent_id = request.GET.get('parent', None)
    prev_id = request.GET.get('prev', None)
    next_id = request.GET.get('next', None)
    if not appmodel or not node_id:
        return JsonResponse(False, safe=False)
    try:
        app_label, model_name = appmodel.split('.')
        model = apps.get_model(app_label=app_label, model_name=model_name)

        # need node ordering for treebeard types
        node_order = getattr(model, 'node_order_by', None)
        node = TreeNode(model.objects.get(pk=node_id))
        if prev_id:
            target = model.objects.get(pk=prev_id)
            node.move(target, 'sorted-sibling' if node_order else 'right')
        elif next_id:
            target = model.objects.get(pk=next_id)
            node.move(target, 'sorted-sibling' if node_order else 'left')
        elif parent_id:
            target = model.objects.get(pk=parent_id)
            node.move(target, 'sorted-child' if node_order else 'first-child')
        else:
            return JsonResponse(False, safe=False)
        return JsonResponse(True, safe=False)
    except:
        return JsonResponse(False, safe=False)



try:
    from django.conf.urls import url
except ImportError:
    # django 4 and up
    from django.urls import re_path as url
from treewidget.views import get_node, move_node

urlpatterns = [
    url(r'get_node/$', get_node, name='treewidget.get_node'),
    url(r'move_node/$', move_node, name='treewidget.move_node'),
]



from django.db.models import QuerySet
from django.db.models import Q, F
from django.db.models.functions import Substr, Length
from django.db.models import CharField, OuterRef, Subquery

try:
    from treebeard.models import Node as TreebeardNode
    from treebeard.al_tree import AL_Node
    from treebeard.ns_tree import NS_Node
    from treebeard.mp_tree import MP_Node
    HAS_TREEBEARD = True
except ImportError:
    HAS_TREEBEARD = False

try:
    from mptt.models import MPTTModel
    HAS_MPTT = True
except ImportError:
    HAS_MPTT = False


TREEBEARD = {
    'root_level': 1,
    'model': {
        'order'     :   lambda model: model.node_order_by or []
    },
    'node': {
        'parent'    :   lambda node: node.get_parent(),
        'prev'      :   lambda node: node.get_prev_sibling(),
        'next'      :   lambda node: node.get_next_sibling(),
        'ancestors' :   lambda node: force_treenode(node.get_ancestors()),
        'descendants':  lambda node: force_treenode(node.get_descendants()),
        'level'     :   lambda node: node.get_depth(),
        'move'      :   lambda node: lambda target, pos: node.move(target, pos),
        'is_root'   :   lambda node: node.is_root(),
    },
}


MPTT = {
    'root_level': 0,
    'model': {
        'order'     :   lambda model: model._mptt_meta.order_insertion_by or []
    },
    'node': {
        'parent'    :   lambda node: getattr(node, node.__class__._mptt_meta.parent_attr, None),
        'prev'      :   lambda node: node.get_previous_sibling(),
        'next'      :   lambda node: node.get_next_sibling(),
        'ancestors' :   lambda node: node.get_ancestors(),
        'descendants':  lambda node: node.get_descendants(),
        'level'     :   lambda node: getattr(node, node.__class__._mptt_meta.level_attr, 0),
        'move'      :   lambda node: lambda target, pos: node.move_to(target, pos),
        'is_root'   :   lambda node: node.is_root_node(),
    },
}


class UnknownTreeImplementation(Exception):
    pass


def get_treetype(model):
    """
    Return the function mapping of the real model tree implementation.
    """
    if HAS_TREEBEARD and issubclass(model, TreebeardNode):
        return TREEBEARD
    elif HAS_MPTT and issubclass(model, MPTTModel):
        return MPTT
    raise UnknownTreeImplementation('cannot map tree implementation')


class TreeQuerySet(object):
    """
    Abstract model tree queryset proxy. It's main purpose is to decorate queryset methods
    in a way, that objects are returned as TreeNode proxy objects for
    common tree attribute access. The real model tree implementation must be known and
    `get_treetype` must return a method mapping for the real tree attributes
    (see `MPTT` and `TREEBEARD` for example definitions).

    The real queryset can be accessed via the `qs` attribute.
    """
    def __init__(self, qs, treetype=None):
        if isinstance(qs, TreeQuerySet):
            self.qs = qs.qs
        else:
            self.qs = qs
        self.treetype = treetype or get_treetype(self.qs.model)

    def __getitem__(self, item):
        item = self.qs[item]
        if item:
            return TreeNode(item, self.qs.model, self.treetype)
        return item

    def _get_next(self):
        for node in self.qs:
            yield TreeNode(node, self.qs.model, self.treetype)

    def __iter__(self):
        for node in self.qs:
            yield TreeNode(node, self.qs.model, self.treetype)

    def __next__(self):
        return next(self)

    def next(self):
        return self.__next__()

    def __getattr__(self, name):
        try:
            return self.__getattribute__(name)
        except AttributeError:
            attr = self.qs.__getattribute__(name)
            if callable(attr):
                return self.proxy(attr)
            return attr

    def proxy(self, attr):
        def f(*args, **kwargs):
            res = attr(*args, **kwargs)
            if isinstance(res, self.qs.__class__):
                return TreeQuerySet(res, self.treetype)
            if isinstance(res, self.qs.model):
                return TreeNode(res, self.qs.model, self.treetype)
            return res
        return f

    @property
    def ordering(self):
        return self.treetype['model']['order'](self.qs.model)

    @property
    def ordering_signs(self):
        return [-1 if attr.startswith('-') else 1 for attr in self.ordering]

    @property
    def appmodel(self):
        return '%s.%s' % (self.qs.model._meta.app_label,
                          self.qs.model._meta.model_name)

    def annotate_parent(self):
        if self.treetype == MPTT:
            parent_field = self.qs.model._mptt_meta.parent_attr
            return TreeQuerySet(self.qs.annotate(_parent_pk=F(parent_field+'__pk')))
        elif self.treetype == TREEBEARD:
            if issubclass(self.qs.model, NS_Node):
                sub = self.qs.model.objects.filter(
                    tree_id=OuterRef('tree_id'),
                    lft__lt=OuterRef('lft'),
                    rgt__gt=OuterRef('rgt')).reverse()[:1]
                qs = self.qs.annotate(_parent_pk=Subquery(sub.values('pk')))
                return TreeQuerySet(qs)
            elif issubclass(self.qs.model, MP_Node):
                sub = self.qs.model.objects.filter(path=OuterRef('parentpath'))
                expr = Substr('path', 1, Length('path') - self.qs.model.steplen,
                              output_field=CharField())
                qs = self.qs.annotate(parentpath=expr).annotate(_parent_pk=Subquery(sub.values('pk')))
                return TreeQuerySet(qs)
            elif issubclass(self.qs.model, AL_Node):
                return TreeQuerySet(
                        self.qs.annotate(_parent_pk=F('parent__pk')))
        raise UnknownTreeImplementation('dont know how to annotate _parent_pk')

    def get_ancestors_parent_annotated(self, include_self=False):
        """
        Creates a queryset containing all parents of the queryset.
        Also annotates the parent pk as `_parent_pk`.
        """
        # django mptt got a ready to go method
        if self.treetype == MPTT:
            parent_field = self.qs.model._mptt_meta.parent_attr
            return TreeQuerySet(
                self.qs.get_ancestors(include_self=include_self)
                    .annotate(_parent_pk=F(parent_field+'__pk')))

        # for treebeard we have to get the parents ourself
        elif self.treetype == TREEBEARD:
            if issubclass(self.qs.model, NS_Node):
                filters = Q()
                for node in self.qs:
                    if include_self:
                        filters |= Q(
                            tree_id=node.tree_id,
                            lft__lte=node.lft,
                            rgt__gte=node.rgt)
                    else:
                        filters |= Q(
                            tree_id=node.tree_id,
                            lft__lt=node.lft,
                            rgt__gt=node.rgt)
                sub = self.qs.model.objects.filter(
                    tree_id=OuterRef('tree_id'),
                    lft__lt=OuterRef('lft'),
                    rgt__gt=OuterRef('rgt')).reverse()[:1]
                qs = self.qs.model.objects.filter(filters)\
                    .annotate(_parent_pk=Subquery(sub.values('pk')))
                return TreeQuerySet(qs)

            elif issubclass(self.qs.model, MP_Node):
                paths = set()
                for node in self.qs:
                    length = len(node.path)
                    if include_self:
                        length += node.steplen
                    paths.update(node.path[0:pos]
                                 for pos in range(node.steplen, length, node.steplen))
                sub = self.qs.model.objects.filter(path=OuterRef('parentpath'))
                expr = Substr('path', 1, Length('path') - self.qs.model.steplen,
                              output_field=CharField())
                qs = self.qs.model.objects.filter(path__in=paths)\
                    .annotate(parentpath=expr)\
                    .annotate(_parent_pk=Subquery(sub.values('pk')))
                return TreeQuerySet(qs)

            elif issubclass(self.qs.model, AL_Node):
                # worst for parent querying
                # we have to walk all levels up to root
                # adds roughly a one query per level
                nodes = self.qs.select_related('parent')
                pks = set()
                parents = set()
                for node in nodes:
                    if include_self:
                        pks.add(node.pk)
                    if node.parent:
                        parents.add(node.parent.pk)
                missing = parents - pks

                while missing:
                    pks.update(parents)
                    parents.clear()
                    for node in self.qs.model.objects.filter(
                            pk__in=missing).select_related('parent'):
                        if node.parent:
                            parents.add(node.parent.pk)
                    missing = parents - pks

                return TreeQuerySet(
                    self.qs.model.objects.filter(pk__in=pks)
                        .annotate(_parent_pk=F('parent__pk')))

        raise UnknownTreeImplementation('dont know how to annotate _parent_pk')


class TreeNode(object):
    """
    Abstract tree node proxy for common tree attribute access.
    The real tree node can be accessed via the `node` attribute.

    NOTE: Only typical tree node attributes get abstracted,
    if you need a specific value from a node, access it via `node`
    (e.g. `obj.node.some_field`).
    """
    def __init__(self, node, model=None, treetype=None):
        self.node = node
        self.model = model or self.node.__class__
        self.treetype = treetype or get_treetype(self.model)

    def _get_real(self, name):
        res = self.treetype['node'][name](self.node)
        if isinstance(res, QuerySet) and res.model == self.model:
            return TreeQuerySet(res, self.treetype)
        if isinstance(res, self.model):
            return TreeNode(res, self.model, self.treetype)
        return res

    def __str__(self):
        return '%s' % self.node

    @property
    def ordering(self):
        return [getattr(self.node, attr.lstrip('-'))
                for attr in self.treetype['model']['order'](self.model)]

    @property
    def parent(self):
        return self._get_real('parent')

    @property
    def prev_sibling(self):
        return self._get_real('prev')

    @property
    def next_sibling(self):
        return self._get_real('next')

    @property
    def ancestors(self):
        return self._get_real('ancestors')

    @property
    def descendants(self):
        return self._get_real('descendants')

    @property
    def level(self):
        return self._get_real('level')

    @property
    def move(self):
        return self._get_real('move')

    @property
    def pk(self):
        return self.node.pk

    @property
    def is_root(self):
        return self._get_real('is_root')


def force_treenode(it):
    """
    Helper function to enforce the content of a returned container
    being TreeNode objects. This especially useful if a manager
    or a queryet method returns a container with tree node objects
    instead of a queryset.
    """
    if isinstance(it, QuerySet):
        return it
    return (TreeNode(node) for node in it)



from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='django-treewidget',
    packages=find_packages(exclude=['example']),
    include_package_data=True,
    version='0.4.1',
    license='MIT',
    description='treewidget for django admin',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Joerg Breitbart',
    author_email='j.breitbart@netzkolchose.de',
    url='https://github.com/jerch/django-treewidget',
    download_url='https://github.com/jerch/django-treewidget/archive/0.4.1.tar.gz',
    keywords=['django', 'widget', 'admin', 'tree'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Database',
        'Topic :: Database :: Front-Ends',
        'Topic :: Software Development :: Libraries',
        'Framework :: Django',
        'Framework :: Django :: 3.2',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
    ],
)
